# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT

import time
import math
from kervi.hal import I2CSensorDeviceDriver

I2C_GYRO_ADDRESS = 0x6A
I2C_ACCL_ADDRESS = 0x1e

MAG_X_MAX = 419
MAG_Y_MAX = 683
MAG_Z_MAX = 528
MAG_X_MIN = -658
MAG_Y_MIN = -397
MAG_Z_MIN = -472

MAG_LPF_FACTOR = 0.4
ACC_LPF_FACTOR = 0.1

RAD_TO_DEG = 57.29578
M_PI = 3.14159265358979323846
G_GAIN = 0.070 # [deg/s/LSB]  If you change the dps for gyro, you need to update this value accordingly
AA = 0.40 # Complementary filter constant
MAG_LPF_FACTOR = 0.4 # Low pass filter constant magnetometer
ACC_LPF_FACTOR = 0.4 # Low pass filter constant for accelerometer
ACC_MEDIANTABLESIZE = 9 # Median filter table size for accelerometer. Higher = smoother but a longer delay
MAG_MEDIANTABLESIZE = 9 # Median filter table size for magnetometer. Higher = smoother but a longer delay

#Kalman filter variables
Q_angle = 0.02
Q_gyro = 0.0015
R_angle = 0.005
y_bias = 0.0
x_bias = 0.0
XP_00 = 0.0
XP_01 = 0.0
XP_10 = 0.0
XP_11 = 0.0
YP_00 = 0.0
YP_01 = 0.0
YP_10 = 0.0
YP_11 = 0.0
KFangleX = 0.0
KFangleY = 0.0

def kalman_filter_y ( accAngle, gyroRate, DT):
	y=0.0
	S=0.0

	global KFangleY
	global Q_angle
	global Q_gyro
	global y_bias
	global YP_00
	global YP_01
	global YP_10
	global YP_11

	KFangleY = KFangleY + DT * (gyroRate - y_bias)

	YP_00 = YP_00 + ( - DT * (YP_10 + YP_01) + Q_angle * DT )
	YP_01 = YP_01 + ( - DT * YP_11 )
	YP_10 = YP_10 + ( - DT * YP_11 )
	YP_11 = YP_11 + ( + Q_gyro * DT )

	y = accAngle - KFangleY
	S = YP_00 + R_angle
	K_0 = YP_00 / S
	K_1 = YP_10 / S
	
	KFangleY = KFangleY + ( K_0 * y )
	y_bias = y_bias + ( K_1 * y )
	
	YP_00 = YP_00 - ( K_0 * YP_00 )
	YP_01 = YP_01 - ( K_0 * YP_01 )
	YP_10 = YP_10 - ( K_1 * YP_00 )
	YP_11 = YP_11 - ( K_1 * YP_01 )
	
	return KFangleY

def kalman_filter_x ( accAngle, gyroRate, DT):
	x=0.0
	S=0.0

	global KFangleX
	global Q_angle
	global Q_gyro
	global x_bias
	global XP_00
	global XP_01
	global XP_10
	global XP_11


	KFangleX = KFangleX + DT * (gyroRate - x_bias)

	XP_00 = XP_00 + ( - DT * (XP_10 + XP_01) + Q_angle * DT )
	XP_01 = XP_01 + ( - DT * XP_11 )
	XP_10 = XP_10 + ( - DT * XP_11 )
	XP_11 = XP_11 + ( + Q_gyro * DT )

	x = accAngle - KFangleX
	S = XP_00 + R_angle
	K_0 = XP_00 / S
	K_1 = XP_10 / S
	
	KFangleX = KFangleX + ( K_0 * x )
	x_bias = x_bias + ( K_1 * x )
	
	XP_00 = XP_00 - ( K_0 * XP_00 )
	XP_01 = XP_01 - ( K_0 * XP_01 )
	XP_10 = XP_10 - ( K_1 * XP_00 )
	XP_11 = XP_11 - ( K_1 * XP_01 )
	
	return KFangleX



class LSM9DS0RawGyroDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, address=I2C_GYRO_ADDRESS, bus=None):
        I2CSensorDeviceDriver.__init__(self, address, bus)
        self.i2c.write8(0x20, 0x0F)
        self.i2c.write_list(0x23, [0x30])

    def read_value(self):

        data0 = self.i2c.read_U8(0x28)
        data1 = self.i2c.read_U8(0x29)

        # Convert the data
        x_gyro = data1 * 256 + data0
        if x_gyro > 32767:
            x_gyro -= 65536

        data0 = self.i2c.read_U8(0x2A)
        data1 = self.i2c.read_U8(0x2B)

        # Convert the data
        y_gyro = data1 * 256 + data0
        if y_gyro > 32767:
            y_gyro -= 65536

        # LSM9DS0 Gyro address, 0x6A(106)
        # Read data back from 0x2C(44), 2 bytes
        # Z-Axis Gyro LSB, Z-Axis Gyro MSB
        data0 = self.i2c.read_U8(0x2C)
        data1 = self.i2c.read_U8(0x2D)

        # Convert the data
        z_gyro = data1 * 256 + data0
        if z_gyro > 32767:
            z_gyro -= 65536

        return (x_gyro, y_gyro, z_gyro)


class LSM9DS0RawAcclDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, address=I2C_ACCL_ADDRESS, bus=None):
        I2CSensorDeviceDriver.__init__(self, address, bus)

        self.i2c.write8(0x20, 0x67)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register2, 0x21(33)
        #		0x20(32)	Full scale = +/-16g
        self.i2c.write8(0x21, 0x20)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register5, 0x24(36)
        #		0x70(112)	Magnetic high resolution, Output data rate = 50Hz
        self.i2c.write8(0x24, 0x70)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register6, 0x25(37)
        #		0x60(96)	Magnetic full scale selection = +/-12 gauss
        self.i2c.write8(0x25, 0x60)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register7, 0x26(38)
        #		0x00(00)	Normal mode, Magnetic continuous conversion mode
        self.i2c.write8(0x26, 0x00)

    def read_value(self):

        data0 = self.i2c.read_U8(0x28)
        data1 = self.i2c.read_U8(0x29)

        # Convert the data
        x_accl = data1 * 256 + data0
        if x_accl > 32767:
            x_accl -= 65536

        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Read data back from 0x2A(42), 2 bytes
        # Y-Axis Accl LSB, Y-Axis Accl MSB
        data0 = self.i2c.read_U8(0x2A)
        data1 = self.i2c.read_U8(0x2B)

        # Convert the data
        y_accl = data1 * 256 + data0
        if y_accl > 32767:
            y_accl -= 65536

        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Read data back from 0x2C(44), 2 bytes
        # Z-Axis Accl LSB, Z-Axis Accl MSB
        data0 = self.i2c.read_U8(0x2C)
        data1 = self.i2c.read_U8(0x2D)

        # Convert the data
        z_accl = data1 * 256 + data0
        if z_accl > 32767:
            z_accl -= 65536

        return (x_accl, y_accl, z_accl)

class LSM9DS0RawMagDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, address=I2C_ACCL_ADDRESS, bus=None):
        I2CSensorDeviceDriver.__init__(self, address, bus)

        self.i2c.write8(0x20, 0x67)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register2, 0x21(33)
        #		0x20(32)	Full scale = +/-16g
        self.i2c.write8(0x21, 0x20)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register5, 0x24(36)
        #		0x70(112)	Magnetic high resolution, Output data rate = 50Hz
        self.i2c.write8(0x24, 0x70)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register6, 0x25(37)
        #		0x60(96)	Magnetic full scale selection = +/-12 gauss
        self.i2c.write8(0x25, 0x60)
        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Select control register7, 0x26(38)
        #		0x00(00)	Normal mode, Magnetic continuous conversion mode
        self.i2c.write8(0x26, 0x00)

    def read_value(self):

        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Read data back from 0x08(08), 2 bytes
        # X-Axis Mag LSB, X-Axis Mag MSB
        data0 = self.i2c.read_U8(0x08)
        data1 = self.i2c.read_U8(0x09)

        # Convert the data
        x_mag = data1 * 256 + data0
        if x_mag > 32767:
            x_mag -= 65536

        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Read data back from 0x0A(10), 2 bytes
        # Y-Axis Mag LSB, Y-Axis Mag MSB
        data0 = self.i2c.read_U8(0x0A)
        data1 = self.i2c.read_U8(0x0B)

        # Convert the data
        y_mag = data1 * 256 + data0
        if y_mag > 32767:
            y_mag -= 65536

        # LSM9DS0 Accl and Mag address, 0x1E(30)
        # Read data back from 0x0C(12), 2 bytes
        # Z-Axis Mag LSB, Z-Axis Mag MSB
        data0 = self.i2c.read_U8(0x0C)
        data1 = self.i2c.read_U8(0x0D)

        # Convert the data
        z_mag = data1 * 256 + data0
        if z_mag > 32767:
            z_mag -= 65536


        return (x_mag, y_mag, z_mag)


class LSM9DS0AcclDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, is_flipped=False, address=I2C_ACCL_ADDRESS, bus=None):
        self.accl = LSM9DS0RawAcclDeviceDriver(address, bus)
        self.mag = LSM9DS0RawMagDeviceDriver(address, bus)


class LSM9DS0OrientationDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, is_flipped=False, accl_address=I2C_ACCL_ADDRESS, gyro_address=I2C_GYRO_ADDRESS, bus=None):
        self.accl = LSM9DS0RawAcclDeviceDriver(accl_address, bus)
        self.mag = LSM9DS0RawMagDeviceDriver(accl_address, bus)
        self.gyro = LSM9DS0RawGyroDeviceDriver(gyro_address, bus)

        self.is_flipped = is_flipped

        self.acc_x_norm = 0.0
        self.acc_y_norm = 0.0
        self.pitch = 0.0
        self.roll = 0.0
        self.mag_x_comp = 0.0
        self.mag_y_comp = 0.0
        self.scaled_mag = [0.0, 0.0, 0.0]

        self.old_x_mag_raw_value = 0
        self.old_y_mag_raw_value = 0
        self.old_z_mag_raw_value = 0
        self.old_x_acc_raw_value = 0
        self.old_y_acc_raw_value = 0
        self.old_z_acc_raw_value = 0

        self.mag_raw = [0, 0, 0]
        self.acc_raw = [0, 0, 0]
        self.gyro_raw = [0, 0, 0]

        self.gyro_x = 0
        self.gyro_y = 0
        self.gyro_z = 0

        self.cf_angle_x = 0
        self.cf_angle_y = 0

        self.last_reading = None

    @property
    def dimensions(self):
        return 3

    @property
    def dimension_labels(self):
        return ["heading","pitch", "roll"]
    
    @property
    def type(self):
        return "orientation"

    @property
    def unit(self):
        return "degree"

    def read_value(self):

        self.mag_raw = list(self.mag.read_value())
        self.acc_raw = list(self.accl.read_value())
        self.gyro_raw = list(self.gyro.read_value())

        rate_gyr_x = self.gyro_raw[0] * G_GAIN
        rate_gyr_y = self.gyro_raw[1] * G_GAIN
        rate_gyr_z = self.gyro_raw[2] * G_GAIN

        acc_x_angle = (math.atan2(self.acc_raw[1],self.acc_raw[2])+math.pi)*RAD_TO_DEG
        acc_y_angle = (math.atan2(self.acc_raw[2],self.acc_raw[0])+math.pi)*RAD_TO_DEG
        
        if not self.last_reading:
            self.gyro_x = rate_gyr_x
            self.gyro_y = rate_gyr_y
            self.gyro_z = rate_gyr_z
        else:
            dt = time.time() - self.last_reading
            self.cf_angle_x = AA * (self.cf_angle_x+rate_gyr_x*dt) + (1 - AA) * acc_x_angle
            self.cf_angle_y = AA * (self.cf_angle_y+rate_gyr_y*dt) + (1 - AA) * acc_y_angle

        self.last_reading = time.time()

        self.mag_raw[0] = self.mag_raw[0] * MAG_LPF_FACTOR + self.old_x_mag_raw_value*(1 - MAG_LPF_FACTOR)
        self.mag_raw[1] = self.mag_raw[1] * MAG_LPF_FACTOR + self.old_y_mag_raw_value*(1 - MAG_LPF_FACTOR)
        self.mag_raw[2] = self.mag_raw[2] * MAG_LPF_FACTOR + self.old_z_mag_raw_value*(1 - MAG_LPF_FACTOR)
        self.acc_raw[0] = self.acc_raw[0] * ACC_LPF_FACTOR + self.old_x_acc_raw_value*(1 - ACC_LPF_FACTOR)
        self.acc_raw[1] = self.acc_raw[1] * ACC_LPF_FACTOR + self.old_y_acc_raw_value*(1 - ACC_LPF_FACTOR)
        self.acc_raw[2] = self.acc_raw[2] * ACC_LPF_FACTOR + self.old_z_acc_raw_value*(1 - ACC_LPF_FACTOR)

        self.old_x_mag_raw_value = self.mag_raw[0]
        self.old_y_mag_raw_value = self.mag_raw[1]
        self.old_z_mag_raw_value = self.mag_raw[2]
        self.old_x_acc_raw_value = self.acc_raw[0]
        self.old_y_acc_raw_value = self.acc_raw[1]
        self.old_z_acc_raw_value = self.acc_raw[2]

        #Apply hard iron calibration
        self.mag_raw[0] -= (MAG_X_MIN + MAG_X_MAX) /2
        self.mag_raw[1] -= (MAG_Y_MIN + MAG_Y_MAX) /2
        self.mag_raw[2] -= (MAG_Z_MIN + MAG_Z_MAX) /2

        #Apply soft iron calibration
        self.scaled_mag[0] = (self.mag_raw[0] - MAG_X_MIN) / (MAG_X_MAX - MAG_X_MIN) * 2 - 1
        self.scaled_mag[1] = (self.mag_raw[1] - MAG_Y_MIN) / (MAG_Y_MAX - MAG_Y_MIN) * 2 - 1
        self.scaled_mag[2] = (self.mag_raw[2] - MAG_Z_MIN) / (MAG_Z_MAX - MAG_Z_MIN) * 2 - 1

        if self.is_flipped:
            self.acc_raw[0] = -self.acc_raw[0]
            self.acc_raw[1] = -self.acc_raw[1]

        #Compute heading
        heading = 180 * math.atan2(self.mag_raw[1], self.mag_raw[0]) / math.pi

        #Convert heading to 0 - 360
        if heading < 0:
            heading += 360

        #Normalize accelerometer raw values.
        accXnorm = self.acc_raw[0] / math.sqrt(self.acc_raw[0] * self.acc_raw[0] + self.acc_raw[1] * self.acc_raw[1] + self.acc_raw[2] * self.acc_raw[2])
        accYnorm = self.acc_raw[1] / math.sqrt(self.acc_raw[0] * self.acc_raw[0] + self.acc_raw[1] * self.acc_raw[1] + self.acc_raw[2] * self.acc_raw[2])

        

        #Calculate pitch and roll
        #pitch = math.asin(accXnorm)
        #roll = -math.asin(accYnorm / math.cos(pitch))

        pitch = self.cf_angle_x
        roll = self.cf_angle_y
        #Calculate the new tilt compensated values
        magXcomp = self.mag_raw[0] * math.cos(pitch) + self.mag_raw[2] * math.sin(pitch)
        magYcomp = self.mag_raw[0] * math.sin(roll) * math.sin(pitch) + self.mag_raw[1] * math.cos(roll) - self.mag_raw[2] * math.sin(roll) * math.cos(pitch)

        #Calculate heading
        heading = 180 * math.atan2(magYcomp, magXcomp)/math.pi

        #Convert heading to 0 - 360
        if heading < 0:
            heading += 360

        return [heading, pitch, roll]

class LSM9DS0GravityDeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, is_flipped=False, address=I2C_ACCL_ADDRESS, bus=None):
        self.accl = LSM9DS0RawAcclDeviceDriver(address, bus)

    @property
    def dimensions(self):
        return 3
    
    @property
    def dimension_labels(self):
        return ["x","y", "z"]
    
    @property
    def type(self):
        return "gravity"

    @property
    def unit(self):
        return "G"
    
    def read_value(self):
        acc_x, acc_y, acc_z = self.accl.read_value()
        return [(acc_x * 0.224)/1000, (acc_y * 0.224)/1000, (acc_z * 0.224)/1000]
    


from kervi import hal
from kervi.devices.gpio.MCP230XX import MCP23017DeviceDriver
from kervi.devices.displays.HD44780 import HD44780RGBDeviceDriver

class CharLCDDeviceDriver(HD44780RGBDeviceDriver):
    """Class to represent and interact with an Adafruit Raspberry Pi character
    LCD plate."""

    def __init__(self, address=0x20, busnum=hal.default_i2c_bus(), cols=16, lines=2):
        """Initialize the character LCD plate.  Can optionally specify a separate
        I2C address or bus number, but the defaults should suffice for most needs.
        Can also optionally specify the number of columns and lines on the LCD
        (default is 16x2).
        """
        # Configure MCP23017 device.
        self._mcp = MCP23017DeviceDriver(address=address, bus=busnum)
        # Set LCD R/W pin to low for writing only.
        self._mcp[LCD_PLATE_RW].define_as_output()
        self._mcp[LCD_PLATE_RW].set(False)
        # Set buttons as inputs with pull-ups enabled.
        for button in (SELECT, RIGHT, DOWN, UP, LEFT):
            self._mcp[button].define_as_input(True)
        # Initialize LCD (with no PWM support).
        super(CharLCDDeviceDriver, self).__init__(
            self._mcp[LCD_PLATE_RS],
            self._mcp[LCD_PLATE_EN],
            self._mcp[LCD_PLATE_D4],
            self._mcp[LCD_PLATE_D5],
            self._mcp[LCD_PLATE_D6],
            self._mcp[LCD_PLATE_D7],
            cols,
            lines,
            self._mcp[LCD_PLATE_RED],
            self._mcp[LCD_PLATE_GREEN],
            self._mcp[LCD_PLATE_BLUE],
            enable_pwm=False
        )

    def is_pressed(self, button):
        """Return True if the provided button is pressed, False otherwise."""
        if button not in set((SELECT, RIGHT, DOWN, UP, LEFT)):
            raise ValueError('Unknown button, must be SELECT, RIGHT, DOWN, UP, or LEFT.')
        return self._mcp.get(button)



# Copyright (c) 2014 Adafruit Industries
# Author: Tony DiCola
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

#This file is a modified version of Tony DiCola's work adapted to the Kervi i2c and sensors api. 


from __future__ import division
import logging
import time
from kervi.hal import I2CSensorDeviceDriver

#BMP085 sensor type
BMP085_TEMPERATURE_SENSOR = 1
BMP085_PRESSURE_SENSOR = 2
BMP085_ALTITUDE_SENSOR = 3

# BMP085 default address.
BMP085_I2CADDR           = 0x77

# Operating Modes
BMP085_ULTRALOWPOWER     = 0
BMP085_STANDARD          = 1
BMP085_HIGHRES           = 2
BMP085_ULTRAHIGHRES      = 3

#private constants
# BMP085 Registers
BMP085_CAL_AC1           = 0xAA  # R   Calibration data (16 bits)
BMP085_CAL_AC2           = 0xAC  # R   Calibration data (16 bits)
BMP085_CAL_AC3           = 0xAE  # R   Calibration data (16 bits)
BMP085_CAL_AC4           = 0xB0  # R   Calibration data (16 bits)
BMP085_CAL_AC5           = 0xB2  # R   Calibration data (16 bits)
BMP085_CAL_AC6           = 0xB4  # R   Calibration data (16 bits)
BMP085_CAL_B1            = 0xB6  # R   Calibration data (16 bits)
BMP085_CAL_B2            = 0xB8  # R   Calibration data (16 bits)
BMP085_CAL_MB            = 0xBA  # R   Calibration data (16 bits)
BMP085_CAL_MC            = 0xBC  # R   Calibration data (16 bits)
BMP085_CAL_MD            = 0xBE  # R   Calibration data (16 bits)
BMP085_CONTROL           = 0xF4
BMP085_TEMPDATA          = 0xF6
BMP085_PRESSUREDATA      = 0xF6

# Commands
BMP085_READTEMPCMD       = 0x2E
BMP085_READPRESSURECMD   = 0x34

class BMP085DeviceDriver(I2CSensorDeviceDriver):
    def __init__(self, sensor_type=BMP085_TEMPERATURE_SENSOR, mode=BMP085_STANDARD, address=BMP085_I2CADDR, bus=0):
        I2CSensorDeviceDriver.__init__(self, address, bus)
        # Check that mode is valid.
        if mode not in [BMP085_ULTRALOWPOWER, BMP085_STANDARD, BMP085_HIGHRES, BMP085_ULTRAHIGHRES]:
            raise ValueError('Unexpected mode value {0}.  Set mode to one of BMP085_ULTRALOWPOWER, BMP085_STANDARD, BMP085_HIGHRES, or BMP085_ULTRAHIGHRES'.format(mode))
        self._mode = mode
        self._sensor_type = sensor_type
        self._load_calibration()

    def _load_calibration(self):
        self.cal_AC1 = self.i2c.read_S16BE(BMP085_CAL_AC1)   # INT16
        self.cal_AC2 = self.i2c.read_S16BE(BMP085_CAL_AC2)   # INT16
        self.cal_AC3 = self.i2c.read_S16BE(BMP085_CAL_AC3)   # INT16
        self.cal_AC4 = self.i2c.read_U16BE(BMP085_CAL_AC4)   # UINT16
        self.cal_AC5 = self.i2c.read_U16BE(BMP085_CAL_AC5)   # UINT16
        self.cal_AC6 = self.i2c.read_U16BE(BMP085_CAL_AC6)   # UINT16
        self.cal_B1 = self.i2c.read_S16BE(BMP085_CAL_B1)     # INT16
        self.cal_B2 = self.i2c.read_S16BE(BMP085_CAL_B2)     # INT16
        self.cal_MB = self.i2c.read_S16BE(BMP085_CAL_MB)     # INT16
        self.cal_MC = self.i2c.read_S16BE(BMP085_CAL_MC)     # INT16
        self.cal_MD = self.i2c.read_S16BE(BMP085_CAL_MD)     # INT16
        

    def _load_datasheet_calibration(self):
        # Set calibration from values in the datasheet example.  Useful for debugging the
        # temp and pressure calculation accuracy.
        self.cal_AC1 = 408
        self.cal_AC2 = -72
        self.cal_AC3 = -14383
        self.cal_AC4 = 32741
        self.cal_AC5 = 32757
        self.cal_AC6 = 23153
        self.cal_B1 = 6190
        self.cal_B2 = 4
        self.cal_MB = -32767
        self.cal_MC = -8711
        self.cal_MD = 2868

    def read_raw_temp(self):
        """Reads the raw (uncompensated) temperature from the sensor."""
        self.i2c.write8(BMP085_CONTROL, BMP085_READTEMPCMD)
        time.sleep(0.005)  # Wait 5ms
        raw = self.i2c.read_U16BE(BMP085_TEMPDATA)
        self.logger.debug('Raw temp 0x{0:X} ({1})', raw & 0xFFFF, raw)
        return raw

    def read_raw_pressure(self):
        """Reads the raw (uncompensated) pressure level from the sensor."""
        self.i2c.write8(BMP085_CONTROL, BMP085_READPRESSURECMD + (self._mode << 6))
        if self._mode == BMP085_ULTRALOWPOWER:
            time.sleep(0.005)
        elif self._mode == BMP085_HIGHRES:
            time.sleep(0.014)
        elif self._mode == BMP085_ULTRAHIGHRES:
            time.sleep(0.026)
        else:
            time.sleep(0.008)
        msb = self.i2c.read_U8(BMP085_PRESSUREDATA)
        lsb = self.i2c.read_U8(BMP085_PRESSUREDATA+1)
        xlsb = self.i2c.read_U8(BMP085_PRESSUREDATA+2)
        raw = ((msb << 16) + (lsb << 8) + xlsb) >> (8 - self._mode)
        self.logger.debug('Raw pressure 0x{0:04X} ({1})', raw & 0xFFFF, raw)
        return raw

    def read_temperature(self):
        """Gets the compensated temperature in degrees celsius."""
        UT = self.read_raw_temp()
        # Datasheet value for debugging:
        #UT = 27898
        # Calculations below are taken straight from section 3.5 of the datasheet.
        X1 = ((UT - self.cal_AC6) * self.cal_AC5) >> 15
        X2 = (self.cal_MC << 11) // (X1 + self.cal_MD)
        B5 = X1 + X2
        temp = ((B5 + 8) >> 4) / 10.0
        self.logger.debug('Calibrated temperature {0} C', temp)
        return temp

    def read_pressure(self):
        """Gets the compensated pressure in Pascals."""
        UT = self.read_raw_temp()
        UP = self.read_raw_pressure()
        # Datasheet values for debugging:
        #UT = 27898
        #UP = 23843
        # Calculations below are taken straight from section 3.5 of the datasheet.
        # Calculate true temperature coefficient B5.
        X1 = ((UT - self.cal_AC6) * self.cal_AC5) >> 15
        X2 = (self.cal_MC << 11) // (X1 + self.cal_MD)
        B5 = X1 + X2
        self.logger.debug('B5 = {0}', B5)
        # Pressure Calculations
        B6 = B5 - 4000
        self.logger.debug('B6 = {0}', B6)
        X1 = (self.cal_B2 * (B6 * B6) >> 12) >> 11
        X2 = (self.cal_AC2 * B6) >> 11
        X3 = X1 + X2
        B3 = (((self.cal_AC1 * 4 + X3) << self._mode) + 2) // 4
        self.logger.debug('B3 = {0}', B3)
        X1 = (self.cal_AC3 * B6) >> 13
        X2 = (self.cal_B1 * ((B6 * B6) >> 12)) >> 16
        X3 = ((X1 + X2) + 2) >> 2
        B4 = (self.cal_AC4 * (X3 + 32768)) >> 15
        self.logger.debug('B4 = {0}', B4)
        B7 = (UP - B3) * (50000 >> self._mode)
        self.logger.debug('B7 = {0}', B7)
        if B7 < 0x80000000:
            p = (B7 * 2) // B4
        else:
            p = (B7 // B4) * 2
        X1 = (p >> 8) * (p >> 8)
        X1 = (X1 * 3038) >> 16
        X2 = (-7357 * p) >> 16
        p = p + ((X1 + X2 + 3791) >> 4)
        self.logger.debug('Pressure {0} Pa', p)
        return p/100

    def read_altitude(self, sealevel_pa=101325.0):
        """Calculates the altitude in meters."""
        # Calculation taken straight from section 3.6 of the datasheet.
        pressure = float(self.read_pressure())
        altitude = 44330.0 * (1.0 - pow(pressure / sealevel_pa, (1.0/5.255)))
        self.logger.debug('Altitude {0} m', altitude)
        return altitude

    def read_sealevel_pressure(self, altitude_m=0.0):
        """Calculates the pressure at sealevel when given a known altitude in
        meters. Returns a value in Pascals."""
        pressure = float(self.read_pressure())
        p0 = pressure / pow(1.0 - altitude_m/44330.0, 5.255)
        self.logger.debug('Sealevel pressure {0} Pa', p0)
        return p0

    def read_value(self):
        if self._sensor_type == BMP085_TEMPERATURE_SENSOR:
            return self.read_temperature()
        elif self._sensor_type == BMP085_PRESSURE_SENSOR:
            return self.read_pressure()
        elif self._sensor_type == BMP085_ALTITUDE_SENSOR:
            return self.read_altitude()

    @property
    def type(self):
        if self._sensor_type == BMP085_TEMPERATURE_SENSOR:
            return "temperature"
        elif self._sensor_type == BMP085_PRESSURE_SENSOR:
            return "pressure"
        elif self._sensor_type == BMP085_ALTITUDE_SENSOR:
            return "altitude"

    @property
    def unit(self):
        if self._sensor_type == BMP085_TEMPERATURE_SENSOR:
            return "C"
        elif self._sensor_type == BMP085_PRESSURE_SENSOR:
            return "hPa"
        elif self._sensor_type == BMP085_ALTITUDE_SENSOR:
            return "m"

    @property
    def max(self):
        if self._sensor_type == BMP085_TEMPERATURE_SENSOR:
            return 85
        elif self._sensor_type == BMP085_PRESSURE_SENSOR:
            return 1100
        elif self._sensor_type == BMP085_ALTITUDE_SENSOR:
            return 9000


    @property
    def min(self):
        if self._sensor_type == BMP085_TEMPERATURE_SENSOR:
            return -40
        elif self._sensor_type == BMP085_PRESSURE_SENSOR:
            return 300
        elif self._sensor_type == BMP085_ALTITUDE_SENSOR:
            return -500



from kervi.hal.motor_controller import MotorControllerBoard, ServoMotorControllerBase, DCMotorControllerBase, StepperMotorControllerBase

class _DummyDCMotorDeviceDriver(DCMotorControllerBase):
    def __init__(self, controller_id, controller_name):
        DCMotorControllerBase.__init__(self, controller_id , controller_name, 4)

    def _set_speed(self, motor, speed):
        print("set speed:", motor, speed)

class _DummyStepperMotorDeviceDriver(StepperMotorControllerBase):
    def __init__(self, controller_id):
        StepperMotorControllerBase.__init__(self, controller_id, "Dummy stepper motor driver", 2)
        self.steps = 0

    def _step(self, motor, style):
        print("step:", motor, style)

    def _release(self, motor):
        print("release step motor:", motor)

    def run(self, step_interval):
        print("stepper, run")


class _DummyServoController(ServoMotorControllerBase):
    def __init__(self, controller_id):
        ServoMotorControllerBase.__init__(self, controller_id, "Dummy servo controller", 4)

    def _set_position(self, channel, position, adjust_min=0, adjust_max=0, adjust_center=0):
        print("servo set position", channel, position)

class DummyMotorBoard(MotorControllerBoard):
    def __init__(self, board_id="dummy_motors", board_name="Dummy motor board"):
        
        MotorControllerBoard.__init__(
            self,
            board_id,
            board_name,
            dc_controller=_DummyDCMotorDeviceDriver(board_id + ".dc_motors", board_name + "-DC motors"),
            stepper_controller=_DummyStepperMotorDeviceDriver(board_id + ".stepper_motors"),
            servo_controller = _DummyServoController(board_id + ".servo_motors"),
        )



from kervi.hal.motor_controller import MotorControllerBoard, DCMotorControllerBase, StepperMotorControllerBase

class _DCMotorController(DCMotorControllerBase):
    def __init__(self, controller_id, ena, in1, in2, enb, in3, in4):
        DCMotorControllerBase.__init__(self, controller_id, "LN298", 2)

        self._motors = [(ena, in1, in2), (enb, in3, in4)]
        for motor in self._motors:
            en, in1, in2 = motor
            en.define_as_pwm()
            in1.define_as_output()
            in2.define_as_output()

    def _set_speed(self, motor, speed):
        """
        Change the speed of a motor on the controller.

        :param motor: The motor to change.
        :type motor: ``int``

        :param speed: Speed from -100 to +100, 0 is stop
        :type speed: ``int``

        """
        self._validate_motor(motor)
        en, in1, in2 = self._motors[motor-1]

        if speed == 0:
            en.pwm_stop()
            in1.set(False)
            in2.set(False)
        elif speed > 0:
            en.pwm_start(abs(speed))
            in1.set(True)
            in2.set(False)
        else:
            en.pwm_start(abs(speed))
            in1.set(False)
            in2.set(True)


class LN298MotorControllerBoard(MotorControllerBoard):
    def __init__(self, ena, in1, in2, enb, in3, in4, board_id="LN298", board_name="LN298"):
        MotorControllerBoard.__init__(
            self,
            board_id,
            board_name,
            dc_controller=_DCMotorController(board_id+".dc_motors", ena, in1, in2, enb, in3, in4)
        )



import time
import kervi.hal as hal
from kervi.hal.gpio import IGPIODeviceDriver
from kervi.devices.gpio.PCF8574 import PCF8574DeviceDriver
from kervi.devices.displays.HD44780 import HD44780DeviceDriver


# Char LCD pin to I/O extender GPIO pin number maps
# PCF8574 pins 4-7,9-12 are bits 0-3,4-7, respectively
# LCD pin name: PCF port bit
PCFPINMAPS = [
{
# map0 - DFRobot, YwRobot, Sainsmart?, also generic black PCB
# http://www.dfrobot.com/image/data/DFR0175/I2C%20LCD%20Backpack%20schematic.pdf
'PCF_RS': 0,
'PCF_RW': 1,
'PCF_EN': 2,
'PCF_BL': 3,
'PCF_D4': 4,
'PCF_D5': 5,
'PCF_D6': 6,
'PCF_D7': 7,
}, {
# map1 - map0 with nibbles swapped
'PCF_RS': 4,
'PCF_RW': 5,
'PCF_EN': 6,
'PCF_BL': 7,
'PCF_D4': 0,
'PCF_D5': 1,
'PCF_D6': 2,
'PCF_D7': 3,
}, {
# map2 - mjkdz board w/22 turn trimmer, GY-LCD-V1
'PCF_RS': 6,
'PCF_RW': 5,
'PCF_EN': 4,
'PCF_BL': 7,
'PCF_D4': 0,
'PCF_D5': 1,
'PCF_D6': 2,
'PCF_D7': 3,
}, {
# map3 - map2 with nibbles swapped
'PCF_RS': 0,
'PCF_RW': 1,
'PCF_EN': 2,
'PCF_BL': 3,
'PCF_D4': 6,
'PCF_D5': 5,
'PCF_D6': 4,
'PCF_D7': 7,
}, {
# map4 - ???
'PCF_RS': 4,
'PCF_RW': 5,
'PCF_EN': 6,
'PCF_BL': 7,
'PCF_D4': 2,
'PCF_D5': 1,
'PCF_D6': 0,
'PCF_D7': 3,
},
]

class CharLCDDeviceDriver(HD44780DeviceDriver):
    """Class to represent and interact with an Adafruit Raspberry Pi character
    LCD plate."""

    def __init__(self, address=0x3f, cols=16, lines=2, pin_map=0, busnum=hal.default_i2c_bus()):
        """Initialize the character LCD plate.  Can optionally specify a separate
        I2C address or bus number, but the defaults should suffice for most needs.
        Can also optionally specify the number of columns and lines on the LCD
        (default is 16x2).
        """
        self._storepinmap(pin_map)
        self._gpio = PCF8574DeviceDriver(address=address, bus=busnum)
        # Set LCD R/W pin to low for writing only.
        self._gpio[PCF_RW].define_as_output()
        self._gpio[PCF_RW].set(False)
               
        super(CharLCDDeviceDriver, self).__init__(
            self._gpio[PCF_RS],
            self._gpio[PCF_EN],
            self._gpio[PCF_D4],
            self._gpio[PCF_D5],
            self._gpio[PCF_D6],
            self._gpio[PCF_D7],
            cols,
            lines,
            backlight=self._gpio[PCF_BL],
            invert_polarity=False,
            enable_pwm=False
        )

    def _storepinmap(self, pinmap):
        try:
            m = PCFPINMAPS[int(pinmap)]
        except TypeError:
            m = pinmap
        validkeys = PCFPINMAPS[0].keys()
        globals().update(dict((k,v) for k,v in m.items() if k in validkeys))





from kervi.hal import get_i2c
from kervi.hal.motor_controller import MotorControllerBoard, DCMotorControllerBase, StepperMotorControllerBase

MOTOR_SPEED_SET = 0x82
PWM_FREQUENCE_SET = 0x84
DIRECTION_SET = 0xaa
MOTOR_SET_A = 0xa1
MOTOR_SET_B = 0xa5
NOTHING = 0x01
ENABLE_STEPPER = 0x1a
UNENABLE_STEPPER = 0x1b
STEPERNU = 0x1c
I2C_MOTOR_DRIVER_ADD = 0x0f #Set the address of the I2CMotorDriver

BOTH_CLOCK_WISE = 0x0a
BOTH_ANTI_CLOCK_WISE = 0x05
M1_CW_M2_ACW = 0x06
M1_ACW_M2CW = 0x09

class _DCMotorDeviceDriver(DCMotorControllerBase):
    def __init__(self, controller_id, address=I2C_MOTOR_DRIVER_ADD, bus=None):
        DCMotorControllerBase.__init__(self, controller_id, "Grove i2c motor driver", 2)
        self.i2c = get_i2c(address, bus)
        self.m1_speed = 0
        self.m2_speed = 0

        self.m1_direction = 1
        self.m2_direction = 1

    def _map(self, x, in_min, in_max, out_min, out_max):
        return abs(int((x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min))

    def _set_speed(self, motor, speed):
        if motor == 1:
            self.m1_speed = self._map(speed, 0, 100, 0, 255)
            if speed >= 0:
                self.m1_direction = 1
            else:
                self.m1_direction = -1

        else:
            self.m2_speed = self._map(speed, 0, 100, 0, 255)
            if speed >= 0:
                self.m2_direction = 1
            else:
                self.m2_direction = -1

        if self.m1_direction == 1 and self.m2_direction == 1:
            direction = BOTH_CLOCK_WISE
        if self.m1_direction == 1 and self.m2_direction == -1:
            direction = M1_CW_M2_ACW
        if self.m1_direction == -1 and self.m2_direction == 1:
            direction = M1_ACW_M2CW
        if self.m1_direction == -1 and self.m2_direction == -1:
            direction = BOTH_ANTI_CLOCK_WISE

        self.i2c.write_list(MOTOR_SPEED_SET, [self.m1_speed, self.m2_speed])
        self.i2c.write_list(DIRECTION_SET, [direction, NOTHING])

class _StepperMotorDeviceDriver(StepperMotorControllerBase):
    def __init__(self, controller_id, address, bus=None):
        StepperMotorControllerBase.__init__(self, controller_id, "Grove i2c motor driver", 2)
        self.i2c = get_i2c(address, bus)

    def step(self, num_step):
        self.i2c.write_i2c_block_data(I2C_MOTOR_DRIVER_ADD, STEPERNU, [num_step, NOTHING])

        ## Enanble the i2c motor driver to drive a 4-wire stepper. the i2c motor driver will
        ## driver a 4-wire with 8 polarity  .
        ## Direction: stepper direction ; 1/0
        ## motor speed: defines the time interval the i2C motor driver change it output to drive the stepper
        ## the actul interval time is : motorspeed * 4ms. that is , when motor speed is 10, the interval time 
        ## would be 40 ms

    def run(self, step_interval):
        if step_interval > 0:
            direction = 1
        else:
            direction = 0

        speed = abs(int(step_interval / 4))
        self.i2c.write_i2c_block_data(I2C_MOTOR_DRIVER_ADD, ENABLE_STEPPER, [direction, speed])

    ##function to uneanble i2C motor drive to drive the stepper.
    def stop(self):
        self.i2c.write_i2c_block_data(I2C_MOTOR_DRIVER_ADD, UNENABLE_STEPPER, [NOTHING, NOTHING])


class GroveMotorController(MotorControllerBoard):
    def __init__(self, address, bus=None, board_id="grove_motor_controller", board_name="Grove motor controller"):
        MotorControllerBoard.__init__(
            self,
            board_id,
            board_name,
            dc_controller=_DCMotorDeviceDriver(board_id + ".dc_motors",address, bus),
            stepper_controller=_StepperMotorDeviceDriver(board_id + ".stepper_motors",address, bus)
        )
